#include <jni.h>
#include <android/log.h>
#include <string>

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <linux/userfaultfd.h>

extern "C"
JNIEXPORT jint JNICALL
Java_com_google_p0_jannh_arthax_MainActivity_nativeMprotectRWX(JNIEnv *env, jobject thiz,
                                                               jlong start, jlong len) {
    if (mprotect((void*)start, len, PROT_READ|PROT_WRITE|PROT_EXEC) == 0)
        return 0;
    return -errno;
}
extern "C"
JNIEXPORT jint JNICALL
Java_com_google_p0_jannh_arthax_MainActivity_nativeUserfaultfd(JNIEnv *env, jobject thiz) {
    __android_log_print(ANDROID_LOG_INFO, "ART-hax", "nativeUserfaultfd(): entry");
    int fd = syscall(__NR_userfaultfd, O_CLOEXEC);
    if (fd == -1)
        return -errno;
    __android_log_print(ANDROID_LOG_INFO, "ART-hax", "nativeUserfaultfd(): syscall success");
    struct uffdio_api api = {
        .api = UFFD_API,
        .features = UFFD_FEATURE_SIGBUS
    };
    __android_log_print(ANDROID_LOG_INFO, "ART-hax", "nativeUserfaultfd(): about to call ioctl");
    if (ioctl(fd, UFFDIO_API, &api)) {
        __android_log_print(ANDROID_LOG_INFO, "ART-hax", "nativeUserfaultfd(): ioctl fail");
        int errno_ = errno;
        close(fd);
        return -errno_;
    }
    __android_log_print(ANDROID_LOG_INFO, "ART-hax", "nativeUserfaultfd(): exit");
    return fd;
}
extern "C"
JNIEXPORT void JNICALL
Java_com_google_p0_jannh_arthax_MainActivity_nativeClose(JNIEnv *env, jobject thiz, jint fd) {
    close(fd);
}

extern "C"
JNIEXPORT jint JNICALL
Java_com_google_p0_jannh_arthax_MainActivity_nativeRegisterUserfault(JNIEnv *env, jobject thiz,
                                                                     jint fd, jlong start,
                                                                     jlong len) {
    struct uffdio_register reg = {
            .range = { .start = (unsigned long)start, .len = (unsigned long)len },
            .mode = UFFDIO_REGISTER_MODE_MISSING
    };
    if (ioctl(fd, UFFDIO_REGISTER, &reg))
        return -errno;
    return 0;
}

extern "C"
JNIEXPORT jint JNICALL
Java_com_google_p0_jannh_arthax_MainActivity_nativeInsertUserfault(JNIEnv *env, jobject thiz,
                                                                   jint fd, jlong addr,
                                                                   jbyteArray data) {
    jboolean isCopy;
    struct uffdio_copy copy = {
            .dst = (unsigned long)addr,
            .src = (unsigned long)env->GetByteArrayElements(data, &isCopy),
            .len = (unsigned long)env->GetArrayLength(data),
            .mode = 0
    };
    int res = ioctl(fd, UFFDIO_COPY, &copy);
    int errno_ = errno;
    env->ReleaseByteArrayElements(data, (jbyte *)copy.src, 0);
    if (res)
        return -errno_;
    return 0;
}